// HACK! Ignore this if used as a top-level compile target.
#ifdef OPEN_LISTS_PROBABILISTIC_OPEN_LIST_H

#include "../globals.h"
#include "../rng.h"
#include "../state_id.h"
#include "../utilities.h"

#include <cmath>
#include <set>

using namespace std;

template<class Entry>
ProbabilisticOpenList<Entry>::ProbabilisticOpenList(const Options &opts)
    : OpenList<Entry>(opts.get<bool>("pref_only")),
      evaluator(opts.get<ScalarEvaluator *>("eval")),
      probability_for_random_g(opts.get<double>("probability_for_random_g")),
      h_probability_base(opts.get<double>("h_probability_base")) {
}

template<class Entry>
ProbabilisticOpenList<Entry>::~ProbabilisticOpenList() {
}

template<class Entry>
Entry ProbabilisticOpenList<Entry>::remove_min(std::vector<int> *key) {
    int max_selectable_g = g_map.size();
    if (g_rng() < probability_for_random_g) {
        max_selectable_g = g_rng(g_map.size());
    }

    double sum = 0;
    int h_min = -1;
    h_min = h_map.begin()->first;
    assert(h_min != -1);

    auto bucket = open_list.end();
    for (auto h_it = h_map.begin(); h_it != h_map.end(); ++h_it) {
        int g_count = 0;
        for (auto g_it = g_map.begin(); g_count <= max_selectable_g; ++g_it, ++g_count) {
            auto tmp_bucket = open_list.find(make_pair(h_it->first, g_it->first));
            if (tmp_bucket != open_list.end()) {
                double prob = pow(h_probability_base, h_it->first - h_min) + numeric_limits<double>::min(); // make sure we don't get 0
                sum += prob;
                if (g_rng() * sum < prob) {
                    bucket = tmp_bucket;
                }
            }
        }
    }
    assert(bucket != open_list.end());

    Entry out = bucket->second.back();
    bucket->second.pop_back();
    auto h_val = h_map.find(bucket->first.first);
    auto g_val = g_map.find(bucket->first.second);
    if (key) {
        assert(key->empty());
        key->push_back(h_val->first);
        key->push_back(g_val->first);
    }
    if (--h_val->second <= 0) {
        assert(h_val->second == 0);
        h_map.erase(h_val);
    }
    if (--g_val->second <= 0) {
        assert(g_val->second == 0);
        g_map.erase(g_val);
    }
    if (bucket->second.empty()) {
        open_list.erase(bucket);
    }
    return out;
}

template<class Entry>
int ProbabilisticOpenList<Entry>::insert(const Entry &entry) {
    if (OpenList<Entry>::only_preferred && !last_preferred) {
        return 0;
    }
    if (dead_end) {
        return 0;
    }
    open_list[make_pair(last_h_value, last_g_value)].push_back(entry);
    ++g_map[last_g_value];
    ++h_map[last_h_value];
    return 1;
}

template<class Entry>
void ProbabilisticOpenList<Entry>::evaluate(int g, bool preferred) {
    evaluator->evaluate(g, preferred);
    last_h_value = evaluator->get_value();
    last_g_value = g;
    last_preferred = preferred;
    dead_end = evaluator->is_dead_end();
    dead_end_reliable = evaluator->dead_end_is_reliable();
}


template<class Entry>
bool ProbabilisticOpenList<Entry>::empty() const {
    return open_list.empty();
}

template<class Entry>
void ProbabilisticOpenList<Entry>::clear() {
    open_list.clear();
}

template<class Entry>
bool ProbabilisticOpenList<Entry>::is_dead_end() const {
    return dead_end;
}

template<class Entry>
bool ProbabilisticOpenList<Entry>::dead_end_is_reliable() const {
    return dead_end_reliable;
}

template<class Entry>
void ProbabilisticOpenList<Entry>::get_involved_heuristics(
    std::set<Heuristic *> &hset) {
    evaluator->get_involved_heuristics(hset);
}

template<class Entry>
OpenList<Entry> *ProbabilisticOpenList<Entry>::_parse(OptionParser &parser) {
    parser.document_synopsis("ProbabilisticOpenList", "");
    parser.add_option<ScalarEvaluator *>("eval", "scalar evaluator");
    parser.add_option<double>("probability_for_random_g", "Probability that a smaller limit for g is selected", "0.5");
    parser.add_option<double>("h_probability_base", "T^(h-h_min)", "0.5");

    parser.add_option<bool>(
        "pref_only",
        "insert only nodes generated by preferred operators", "false");

    Options opts = parser.parse();

    if (!in_range(opts.get<double>("probability_for_random_g"), 0.0, 1.1)) {
        parser.error("probability_for_random_g must be between 0 and 1.");
    }
    if (!in_range(opts.get<double>("h_probability_base"), 0.0, 1.1)) {
        parser.error("h_probability_base must be between 0 and 1.");
    }
    if (parser.dry_run())
        return 0;
    else
        return new ProbabilisticOpenList<Entry>(opts);
}

#endif
