#ifndef RANDOM_WALK_SEARCH_H
#define RANDOM_WALK_SEARCH_H

#include "search_engine.h"
#include "search_space.h"

#include <cmath>
#include <unordered_set>
#include <vector>

class Heuristic;
class GlobalOperator;
class ScalarEvaluator;
class Options;
class RandomWalkType;

/*
  Search engine based on Monte-Carlo random walks, known from ARVAND.
  Only the end point of a random path is evaluated.
  Based on:
    Hootan Nakhost and Martin Mueller.  Monte-carlo exploration for
    deterministic planning. InProceedings of the 21st International
    Joint Conference on Artificial Intelligence (IJCAI 2009),
    pages 1766-1771, 2009.
*/
class RandomWalkSearch : public SearchEngine {
public:
    typedef std::vector<std::pair<GlobalState, const GlobalOperator *> > Path;

    enum WalkResult {
        NODE,
        GOAL,
        DEAD_END
    };

private:
    RandomWalkType *walk_type;
    Path global_path;
    std::unordered_set<StateID, __gnu_cxx::hash<StateID> > visited;

    const int max_explorations_without_improvement;
    const int walks_per_exploration;
    const int default_walk_length;

    Heuristic *heuristic;

    const int extending_period;
    const double extending_rate;

    const double weight_for_acceptable_progress;
    bool use_acceptable_progress;
    double last_acceptable_progress;

    GlobalState current_state;
    int h_min;
    int init_h;
    int num_explorations_without_improvement;
    int num_resets;

    double compute_acceptable_progress(int progress) {
        return (1 - weight_for_acceptable_progress) * last_acceptable_progress + weight_for_acceptable_progress * progress;
    }

protected:
    WalkResult do_single_monte_carlo_random_walk(const GlobalState &root, Path &tmp_path, const int walk_length);
    Path do_monte_carlo_exploration(const GlobalState &root);

    void reset();
    virtual SearchStatus step();

    virtual void initialize();

    void build_plan_in_search_space();

public:
    explicit RandomWalkSearch(const Options &opts);
    virtual ~RandomWalkSearch();

    void statistics() const;
    void update_global_path(const Path &global_path);
};


class RandomWalkType {
public:
    RandomWalkType();
    virtual ~RandomWalkType();
    virtual void reset() {}
    virtual const GlobalOperator *choose_operator(const std::vector<const GlobalOperator *> &applicable_ops) = 0;
    virtual void handle_walk(const RandomWalkSearch::Path &, RandomWalkSearch::WalkResult) {}
    virtual void handle_progress(Heuristic *) {}
};

class PureWalk : public RandomWalkType {
public:
    PureWalk();
    virtual ~PureWalk();
    virtual const GlobalOperator *choose_operator(const std::vector<const GlobalOperator *> &applicable_ops);
};

class ProbabilisticWalk : public RandomWalkType {
    double tau;
protected:
    std::vector<double> q_values;

public:
    ProbabilisticWalk(double tau);
    virtual ~ProbabilisticWalk();
    virtual void reset();
    virtual const GlobalOperator *choose_operator(const std::vector<const GlobalOperator *> &applicable_ops);
};

class MDAWalk : public ProbabilisticWalk {
    std::vector<int> num_operator_used_in_failed_walks;
    std::vector<int> num_operator_used_in_successful_walks;

public:
    MDAWalk(const Options &opts);
    virtual void reset();
    virtual void handle_walk(const RandomWalkSearch::Path &path, RandomWalkSearch::WalkResult walk_result);
};

class MHAWalk : public ProbabilisticWalk {
public:
    MHAWalk(const Options &opts);
    virtual void handle_progress(Heuristic *heuristic);
};

#endif
